/*
 * Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/*
 * based in parts on libfaad, Copyright (C) 2003-2005 M. Bakker,
 * Ahead Software AG
 */

/** \file aac_afh.c para_server's aac audio format handler. */

#include <regex.h>
#include <neaacdec.h>

#include "para.h"

/* To get the mp4ff_tag_t and mp4ff_metadata_t typedefs. */
#define USE_TAGGING
#include <mp4ff.h>

#include "error.h"
#include "portable_io.h"
#include "afh.h"
#include "string.h"
#include "fd.h"


struct aac_afh_context {
	const void *map;
	size_t mapsize;
	size_t fpos;
	int32_t track;
	mp4ff_t *mp4ff;
	mp4AudioSpecificConfig masc;
	mp4ff_callback_t cb;
};

static uint32_t aac_afh_read_cb(void *user_data, void *dest, uint32_t want)
{
	struct aac_afh_context *c = user_data;
	uint32_t have, rv;

	if (want == 0 || c->fpos >= c->mapsize) {
		PARA_INFO_LOG("failed attempt to read %u bytes @%zu\n", want,
			c->fpos);
		errno = EAGAIN;
		return -1;
	}
	have = c->mapsize - c->fpos;
	rv = PARA_MIN(have, want);
	PARA_DEBUG_LOG("reading %u bytes @%zu\n", rv, c->fpos);
	memcpy(dest, c->map + c->fpos, rv);
	c->fpos += rv;
	return rv;
}

static uint32_t aac_afh_seek_cb(void *user_data, uint64_t pos)
{
	struct aac_afh_context *c = user_data;
	c->fpos = pos;
	return 0;
}

static int32_t aac_afh_get_track(mp4ff_t *mp4ff, mp4AudioSpecificConfig *masc)
{
	int32_t i, rc, num_tracks = mp4ff_total_tracks(mp4ff);

	assert(num_tracks >= 0);
	for (i = 0; i < num_tracks; i++) {
		unsigned char *buf = NULL;
		unsigned buf_size = 0;

		mp4ff_get_decoder_config(mp4ff, i, &buf, &buf_size);
		if (buf) {
			rc = NeAACDecAudioSpecificConfig(buf, buf_size, masc);
			free(buf);
			if (rc < 0)
				continue;
			return i;
		}
	}
	return -1; /* no audio track */
}

static int aac_afh_open(const void *map, size_t mapsize, void **afh_context)
{
	int ret;
	struct aac_afh_context *c = para_malloc(sizeof(*c));

	c->map = map;
	c->mapsize = mapsize;
	c->fpos = 0;
	c->cb.read = aac_afh_read_cb;
	c->cb.seek = aac_afh_seek_cb;
	c->cb.user_data = c;

	ret = -E_MP4FF_OPEN;
	c->mp4ff = mp4ff_open_read(&c->cb);
	if (!c->mp4ff)
		goto free_ctx;
	c->track = aac_afh_get_track(c->mp4ff, &c->masc);
	ret = -E_MP4FF_TRACK;
	if (c->track < 0)
		goto close_mp4ff;
	*afh_context = c;
	return 0;
close_mp4ff:
	mp4ff_close(c->mp4ff);
free_ctx:
	free(c);
	*afh_context = NULL;
	return ret;
}

static void aac_afh_close(void *afh_context)
{
	struct aac_afh_context *c = afh_context;
	mp4ff_close(c->mp4ff);
	free(c);
}

/**
 * Libmp4ff function to reposition the file to the given sample.
 *
 * \param f The opaque handle returned by mp4ff_open_read().
 * \param track The number of the (audio) track.
 * \param sample Destination.
 *
 * We need this function to obtain the offset of the sample within the audio
 * file. Unfortunately, it is not exposed in the mp4ff header.
 *
 * \return This function always returns 0.
 */
int32_t mp4ff_set_sample_position(mp4ff_t *f, const int32_t track, const int32_t sample);

static int aac_afh_get_chunk(long unsigned chunk_num, void *afh_context,
		const char **buf, size_t *len)
{
	struct aac_afh_context *c = afh_context;
	int32_t ss;
	size_t offset;

	assert(chunk_num <= INT_MAX);
	/* this function always returns zero */
	mp4ff_set_sample_position(c->mp4ff, c->track, chunk_num);
	offset = c->fpos;
	ss = mp4ff_read_sample_getsize(c->mp4ff, c->track, chunk_num);
	if (ss <= 0)
		return -E_MP4FF_BAD_SAMPLE;
	assert(ss + offset <= c->mapsize);
	*buf = c->map + offset;
	*len = ss;
	return 1;
}

static void _aac_afh_get_taginfo(const mp4ff_t *mp4ff, struct taginfo *tags)
{
	mp4ff_meta_get_artist(mp4ff, &tags->artist);
	mp4ff_meta_get_title(mp4ff, &tags->title);
	mp4ff_meta_get_date(mp4ff, &tags->year);
	mp4ff_meta_get_album(mp4ff, &tags->album);
	mp4ff_meta_get_comment(mp4ff, &tags->comment);
}

/*
 * Init m4a file and write some tech data to given pointers.
 */
static int aac_get_file_info(char *map, size_t numbytes, __a_unused int fd,
		struct afh_info *afhi)
{
	int ret;
	int32_t rv;
	struct aac_afh_context *c;
	int64_t tmp;
	const char *buf;
	size_t sz;
	uint32_t n;

	ret = aac_afh_open(map, numbytes, (void **)&c);
	if (ret < 0)
		return ret;

	ret = -E_MP4FF_BAD_SAMPLERATE;
	rv = mp4ff_get_sample_rate(c->mp4ff, c->track);
	if (rv <= 0)
		goto close;
	afhi->frequency = rv;

	ret = -E_MP4FF_BAD_CHANNEL_COUNT;
	rv = mp4ff_get_channel_count(c->mp4ff, c->track);
	if (rv <= 0)
		goto close;
	afhi->channels = rv;

	ret = -E_MP4FF_BAD_SAMPLE_COUNT;
	rv = mp4ff_num_samples(c->mp4ff, c->track);
	if (rv <= 0)
		goto close;
	afhi->chunks_total = rv;
	afhi->max_chunk_size = 0;
	for (n = 0; n < afhi->chunks_total; n++) {
		if (aac_afh_get_chunk(n, c, &buf, &sz) < 0)
			break;
		afhi->max_chunk_size = PARA_MAX((size_t)afhi->max_chunk_size, sz);
	}

	tmp = c->masc.sbr_present_flag == 1? 2048 : 1024;
	afhi->seconds_total = tmp * afhi->chunks_total / afhi->frequency;
	ms2tv(1000 * tmp / afhi->frequency, &afhi->chunk_tv);

	if (aac_afh_get_chunk(0, c, &buf, &sz) >= 0)
		numbytes -= buf - map;
	afhi->bitrate = 8 * numbytes / afhi->seconds_total / 1000;
	_aac_afh_get_taginfo(c->mp4ff, &afhi->tags);
	ret = 1;
close:
	aac_afh_close(c);
	return ret;
}

static uint32_t aac_afh_meta_read_cb(void *user_data, void *dest, uint32_t want)
{
	int fd = *(int *)user_data;
	return read(fd, dest, want);
}

static uint32_t aac_afh_meta_seek_cb(void *user_data, uint64_t pos)
{
	int fd = *(int *)user_data;
	return lseek(fd, pos, SEEK_SET);
}

static uint32_t aac_afh_meta_write_cb(void *user_data, void *dest, uint32_t want)
{
	int fd = *(int *)user_data;
	return write(fd, dest, want);
}

static uint32_t aac_afh_meta_truncate_cb(void *user_data)
{
	int fd = *(int *)user_data;
	off_t offset = lseek(fd, 0, SEEK_CUR);
	return ftruncate(fd, offset);
}

static void replace_tag(mp4ff_tag_t *tag, const char *new_val, bool *found)
{
	free(tag->value);
	tag->value = para_strdup(new_val);
	*found = true;
}

static void add_tag(mp4ff_metadata_t *md, const char *item, const char *value)
{
	md->tags[md->count].item = para_strdup(item);
	md->tags[md->count].value = para_strdup(value);
	md->count++;
}

static int aac_afh_rewrite_tags(const char *map, size_t mapsize,
		struct taginfo *tags, int fd, __a_unused const char *filename)
{
	int ret, i;
	int32_t rv;
	mp4ff_metadata_t metadata;
	mp4ff_t *mp4ff;
	mp4ff_callback_t cb = {
		.read = aac_afh_meta_read_cb,
		.seek = aac_afh_meta_seek_cb,
		.write = aac_afh_meta_write_cb,
		.truncate = aac_afh_meta_truncate_cb,
		.user_data = &fd
	};
	bool found_artist = false, found_title = false, found_album = false,
		found_year = false, found_comment = false;

	ret = write_all(fd, map, mapsize);
	if (ret < 0)
		return ret;
	lseek(fd, 0, SEEK_SET);

	mp4ff = mp4ff_open_read_metaonly(&cb);
	if (!mp4ff)
		return -E_MP4FF_OPEN;

	ret = -E_MP4FF_META_READ;
	rv = mp4ff_meta_get_num_items(mp4ff);
	if (rv < 0)
		goto close;
	metadata.count = rv;
	PARA_NOTICE_LOG("%d metadata item(s) found\n", rv);

	metadata.tags = para_malloc((metadata.count + 5) * sizeof(mp4ff_tag_t));
	for (i = 0; i < metadata.count; i++) {
		mp4ff_tag_t *tag = metadata.tags + i;

		ret = -E_MP4FF_META_READ;
		if (mp4ff_meta_get_by_index(mp4ff, i,
				&tag->item, &tag->value) < 0)
			goto free_tags;
		PARA_INFO_LOG("found: %s: %s\n", tag->item, tag->value);
		if (!strcmp(tag->item, "artist"))
			replace_tag(tag, tags->artist, &found_artist);
		else if (!strcmp(tag->item, "title"))
			replace_tag(tag, tags->title, &found_title);
		else if (!strcmp(tag->item, "album"))
			replace_tag(tag, tags->album, &found_album);
		else if (!strcmp(tag->item, "date"))
			replace_tag(tag, tags->year, &found_year);
		else if (!strcmp(tag->item, "comment"))
			replace_tag(tag, tags->comment, &found_comment);
	}
	if (!found_artist)
		add_tag(&metadata, "artist", tags->artist);
	if (!found_title)
		add_tag(&metadata, "title", tags->title);
	if (!found_album)
		add_tag(&metadata, "album", tags->album);
	if (!found_year)
		add_tag(&metadata, "date", tags->year);
	if (!found_comment)
		add_tag(&metadata, "comment", tags->comment);
	ret = -E_MP4FF_META_WRITE;
	if (mp4ff_meta_update(&cb, &metadata) < 0)
		goto free_tags;
	ret = 1;
free_tags:
	for (; i > 0; i--) {
		free(metadata.tags[i - 1].item);
		free(metadata.tags[i - 1].value);
	}
	free(metadata.tags);
close:
	mp4ff_close(mp4ff);
	return ret;
}

static const char * const aac_suffixes[] = {"m4a", "mp4", NULL};
/**
 * the init function of the aac audio format handler
 *
 * \param afh pointer to the struct to initialize
 */
void aac_afh_init(struct audio_format_handler *afh)
{
	afh->get_file_info = aac_get_file_info,
	afh->suffixes = aac_suffixes;
	afh->rewrite_tags = aac_afh_rewrite_tags;
	afh->open = aac_afh_open;
	afh->get_chunk = aac_afh_get_chunk;
	afh->close = aac_afh_close;
}
